/* server.c -	 a server program that uses the socket interface to tcp
 *		 This server serves multiple clients at a time.
 *		 For each client it creates a child process that
 *		 repeatedly reads a string,
 *		 and then sends it back to the client
 *		 Modified for exs API.
 *  on Solaris this program must be linked with -lnsl and -lsocket
 */

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE

#if defined(__sun__)
	#define __EXTENSIONS__
#endif

#define _BSD_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <errno.h>
#include <arpa/inet.h>
#include <sys/socket.h>
#include <ctype.h>
#include <string.h>
#include <sys/stat.h>
#include <sys/time.h>
#include <sys/resource.h>
#include <fcntl.h>
#include <time.h>
#include <netdb.h>

#include <sys/types.h>
#include <sys/wait.h>
#include <signal.h>
#include <pthread.h>

#include "utilities.h"
#include "common.h"


struct child_params {
	int			client_fd;
	struct sockaddr_storage	client_addr;
	};


void *
server_thread(void *arg)
{
	int			len, client_fd, client_port, i;
	size_t			mess_size;
	int			count1, count2;
	char			client_ip[TEXT_SIZE];
	char			text_buf[TEXT_SIZE];
	char			*in_ptr;
	struct sockaddr		*client_addr;
	struct child_params	*param;
	struct sockaddr_storage	name;
	socklen_t		namelen;
	exs_event_t		completion;
	struct rusage		start_usage;
	int			secs, usecs;
	int			user_secs, user_usecs;
	int			kernel_secs, kernel_usecs;


	param = (struct child_params *)arg;
	client_fd = param->client_fd;
	client_addr = (struct sockaddr *)&param->client_addr;

	strcpy(client_ip, "0.0.0.0");
	client_port = 0;
	
	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu connected on "
			"fd %d, client %p", time_prefix(), pthread_self(),
			client_fd, client_addr);

	/* we are now successfully connected to a remote client */
#if defined(ALLOW_DEBUG)
	if (get_debug_mask() & LOG_APPLICATION) {
		if (client_addr != NULL) {
			/* can we get address info about remote client */
			print_ip_address((struct sockaddr *)&param->client_addr,
					"remote client",
					"latenexs");
		}
	}
#endif

	/* try out the added exs_getsockname() and exs_getpeername() functions*/
	namelen = sizeof(name);
	if (exs_getsockname(client_fd, (struct sockaddr *)&name, &namelen))
		perror("exs_getsockname");
	print_ip_address((struct sockaddr *)&name, "server exs_getsockname",
								"latenexs");

	namelen = sizeof(name);
	if (exs_getpeername(client_fd, (struct sockaddr *)&name, &namelen))
		perror("exs_getpeername");
	print_ip_address((struct sockaddr *)&name, "server exs_getpeername",
								"latenexs");

	/* repeatedly read a message from a client,
	 * then write it to standard output along with the client's id
	 */
	len = size;
	mess_size = size;
	count1 = count - 1;
	count2 = count1 - 1;

        /* wait for client to tell us to "go" */
        len = readpacket(client_fd, text_buf, TEXT_SIZE,
                                        EXS_MHANDLE_UNREGISTERED, recv_queue);
	if (len <= 0) {
		fprintf(stderr,"server: server_thread readpacket %s\n",strerror(errno));
		return NULL;
	}
	fprintf(stderr, "server: readpacket \"%s\"\n", text_buf);



	/* set up to receive first "ping" from client */
	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu calling first "
			"exs_recv fd %d in_mess %p len %zu recv_queue %p flags "
			"%0x", time_prefix(), pthread_self(), client_fd,
			in_mess, mess_size, recv_queue, recv_flags);
	if (exs_recv(client_fd, in_mess, mess_size, recv_flags,
				recv_queue, NULL, in_handle) < 0) {
		fprintf(stderr, "%s server exs_recv: %s\n",
					time_prefix(), strerror(errno));
		goto err2;
	}

	/* send an ack back to client to say we're ready */
        strcpy(text_buf, "ok");
	fprintf(stderr, "server: writeblock \"%s\"\n", text_buf);
	len = writeblock(client_fd, text_buf, 3, 0,
					EXS_MHANDLE_UNREGISTERED, send_queue);
	if (len <= 0) {
		fprintf(stderr,"server: server_thread writeblock %s\n",
			strerror(errno));
		return NULL;
	}

	/* now wait for first "ping" to actually arrive from client */
	if (exs_qdequeue(recv_queue, &completion, 1, NULL) < 0) {
		fprintf(stderr, "%s server recv exs_qdequeue: %s\n",
					time_prefix(), strerror(errno));
		goto err2;
	}

	TRACE_DBG(LOG_APPLICATION, "server: first ping dequeued, start timers");

	/* now start the timers */
	if( getrusage(RUSAGE_SELF, &start_usage) < 0 )
		{
		printf("server: getrusage start %s\n",
					strerror(errno));
		exit(EXIT_FAILURE);
		}
	if (clock_gettime(CLOCK_EXS, &start_time) < 0)
		{
		fprintf(stderr,
			"%s server: clock_gettime start %s\n",
			time_prefix(), strerror(errno));
		exit(EXIT_FAILURE);
		}

	for (i = 0; i < count1; i++) {

		TRACE_DBG(LOG_APPLICATION,"%s server: starting iteration %d",
				time_prefix(), i);

		in_ptr = completion.exs_evt_union.exs_evt_xfer.exs_evt_buffer;

#if defined(ALLOW_FULL_CHECKING)
		if (completion.exs_evt_errno != 0) {
			fprintf(stderr, "%s server recv completion: %s\n",
					time_prefix(),
					strerror(completion.exs_evt_errno));
			break;
		}

		len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;

		TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu "
				"exs_recv returned len %d",
				time_prefix(), pthread_self(), len);
		if (len <= 0)
			break;

		TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu read "
				"len %d", time_prefix(), pthread_self(), len);

#if defined(ALLOW_DEBUG)
		if (get_debug_mask() & LOG_APPLICATION) {
			/* make it a string for printing */
			in_ptr[len] = '\0';
			TRACE_DBG(LOG_APPLICATION,
				"%s server: child thread_id %lu recvd "
				"%d char string: \"%.40s\"", time_prefix(),
				pthread_self(), len, in_ptr);

			TRACE_DBG(LOG_APPLICATION,
				"%s server: child thread_id %lu sending"
				" %d char string \"%.40s\"", time_prefix(),
				pthread_self(), len, in_ptr);
		}
#endif
#endif

#if 0
		TRACE_DBG(LOG_APPLICATION, "%s server: sleeping 1 second",
				time_prefix());
		sleep(1);
#endif

		/* send "pong" reply to "ping" just received (without event) */
		TRACE_DBG(LOG_APPLICATION, "%s server: send pong reply \"%.40s\","
				" mess_size %zu", time_prefix(),
				in_mess, mess_size);
		if (exs_send(client_fd, in_mess, mess_size, EXS_UNSIGNALED,
				NULL, NULL, in_handle) < 0) {
			fprintf(stderr, "%s server exs_send: %s\n",
						time_prefix(), strerror(errno));
			break;
		}

#if 0
		/* wait for "pong" reply to finish */
		if (exs_qdequeue(send_queue, &completion, 1, NULL) < 0) {
			fprintf(stderr, "%s server send exs_qdequeue: %s\n",
						time_prefix(), strerror(errno));
			break;
		}

#if defined(ALLOW_FULL_CHECKING)
		if (completion.exs_evt_errno != 0) {
			fprintf(stderr, "%s server send completion: %s\n",
					time_prefix(),
					strerror(completion.exs_evt_errno));
			break;
		}

		if (completion.exs_evt_union.exs_evt_xfer.exs_evt_length
								!= mess_size)
			break;
#endif
#endif

		/* set up to receive next "ping" from client */
		TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu calling"
				" exs_recv fd %d in_mess %p len %zu recv_queue "
				"%p flags %0x", time_prefix(), pthread_self(),
				client_fd, in_mess, mess_size, recv_queue,
				recv_flags);
		if (exs_recv(client_fd, in_mess, mess_size, recv_flags,
					recv_queue, NULL, in_handle) < 0) {
			fprintf(stderr, "%s server exs_recv: %s\n",
						time_prefix(), strerror(errno));
			break;
		}

		/* now wait for next "ping" to actually arrive from client */
		TRACE_DBG(LOG_APPLICATION, "%s server: wait for next ping",
				time_prefix());
		if (exs_qdequeue(recv_queue, &completion, 1, NULL) < 0) {
			fprintf(stderr, "%s server recv exs_qdequeue: %s\n",
						time_prefix(), strerror(errno));
			break;
		}
	}	/* for loop */

	TRACE_DBG(LOG_APPLICATION, "%s server: starting final iteration %d",
							time_prefix(), i);

	in_ptr = completion.exs_evt_union.exs_evt_xfer.exs_evt_buffer;

#if defined(ALLOW_FULL_CHECKING)
	if (completion.exs_evt_errno != 0) {
		fprintf(stderr, "%s server final recv completion: %s\n",
				time_prefix(),
				strerror(completion.exs_evt_errno));
		goto err1;
	}

	len = completion.exs_evt_union.exs_evt_xfer.exs_evt_length;

	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu "
			"final exs_recv returned len %d",
			time_prefix(), pthread_self(), len);
	if (len <= 0) {
		fprintf(stderr, "server final recv exs_evt_length %d <= 0\n",
			len);
		goto err1;
	}

	TRACE_DBG(LOG_APPLICATION, "%s server: child thread_id %lu final read "
			"len %d", time_prefix(), pthread_self(), len);

#if defined(ALLOW_DEBUG)
	if (get_debug_mask() & LOG_APPLICATION) {
		/* make it a string for printing */
		in_ptr[len] = '\0';
		TRACE_DBG(LOG_APPLICATION,
			"%s server: child thread_id %lu recvd "
			"%d char string: \"%.40s\"", time_prefix(),
			pthread_self(), len, in_ptr);

		TRACE_DBG(LOG_APPLICATION,
			"%s server: child thread_id %lu sending"
			" final %d char string \"%.40s\"", time_prefix(),
			pthread_self(), len, in_ptr);
	}
#endif
#endif

	TRACE_DBG(LOG_APPLICATION, "%s server: sending final pong reply, "
					"iteration %d", time_prefix(), i);

	/* send final "pong" reply to final "ping" just received */
	TRACE_DBG(LOG_APPLICATION, "%s server: exs_send final pong reply, "
				"size %zu, iteration %d", time_prefix(),
				mess_size, i);
	if (exs_send(client_fd, in_mess, mess_size, 0,
				send_queue, NULL, in_handle) < 0) {
		fprintf(stderr, "%s server final exs_send: %s\n",
					time_prefix(), strerror(errno));
		goto err1;
	}

	TRACE_DBG(LOG_APPLICATION, "%s server: waiting for final exs_send, "
				"iteration %d", time_prefix(), i);

	/* wait for final "pong" reply to finish */
	if (exs_qdequeue(send_queue, &completion, 1, NULL) < 0) {
		fprintf(stderr, "%s server send final exs_qdequeue: %s\n",
					time_prefix(), strerror(errno));
		goto err1;
	}

#if defined(ALLOW_FULL_CHECKING)
	if (completion.exs_evt_errno != 0) {
		fprintf(stderr, "%s server final send completion: %s\n",
				time_prefix(),
				strerror(completion.exs_evt_errno));
	}

	if (completion.exs_evt_union.exs_evt_xfer.exs_evt_length != mess_size) {
		fprintf(stderr, "server final send exs_evt_length %ld != "
			"mess_size %zu\n",
			completion.exs_evt_union.exs_evt_xfer.exs_evt_length,
			mess_size);
	}

#endif

	/* don't forget to count the last message */
	i++;
err1:

	TRACE_DBG(LOG_APPLICATION, "%s server: stop timers, iteration %d",
				time_prefix(), i);

	/* stop timers immediately after last transaction completes */
	get_relative_time(&start_time, &secs, &usecs, "server");
	get_relative_usage(&start_usage, &user_secs, &user_usecs,
					&kernel_secs, &kernel_usecs, "server");

	print_latency("server", i, secs, usecs);
	fprintf(stderr, "server: %11d.%06d seconds elapsed\n", secs, usecs);
        fprintf(stderr, "server: %11d.%06d seconds user (%.4f%%)\n",
			user_secs, user_usecs,
			percentage(secs, usecs, user_secs, user_usecs));
	fprintf(stderr, "server: %11d.%06d seconds kernel (%.4f%%)\n",
			kernel_secs, kernel_usecs,
			percentage(secs, usecs, kernel_secs, kernel_usecs));

err2:
	if (use_registered_memory) {
		/* deregister the RDMA memory buffers */
		if(exs_mderegister(out_handle, 0)) {
			perror("child exs_mderegister out memory");
		}
		if(exs_mderegister(in_handle, 0)) {
			perror("child exs_mderegister  in memory");
		}
	}

	/* give client time to get the reply */
	sleep(1);

	/* close the connection to the client */
	TRACE_DBG(LOG_APPLICATION,
			"%s server: child pthread_id %lu disconnecting"
			" from client at IP address %s port %d",
			time_prefix(), pthread_self(), client_ip, client_port);
	if (exs_close(client_fd, 0, recv_queue, NULL) < 0) {
		perror("child exs_close connection to client");
	} else {
		if (asynch_wait(recv_queue, &completion, NULL) < 0) {
			perror("serverexs: server exs_close asynch_wait");
		}
	}

	if (exs_qdelete(recv_queue) < 0) {
		perror("child exs_qdelete recv_queue");
	}
	if (exs_qdelete(send_queue) < 0) {
		perror("child exs_qdelete send_queue");
	}
	//pthread_exit(NULL);
	return NULL;
}	/* server_thread */

int
main(int argc, char *argv[])
{
	int			listening_fd, nevents;
	exs_qhandle_t		accept_queue;
	struct sockaddr_storage	server_addr;
	exs_acceptaddr_t	address_vector[ADDRESS_ITEMS];
	exs_event_t		accept_vector[ACCEPT_ITEMS];
	struct child_params	param;


	startup(argc, argv, 0);


	/* need to create an event_queue for accepting connects to server */
	accept_queue = exs_qcreate(ACCEPT_DEPTH);
	if (accept_queue == EXS_QHANDLE_INVALID) {
		perror("server exs_qcreate");
		exit(EXIT_FAILURE);
	}


	/* establish a server "listening post" */
	TRACE_DBG(LOG_APPLICATION, "%s server: calling openserver to "
			"establish listening post", time_prefix());
	listening_fd = openserver(server_port, server_name, ACCEPT_DEPTH,
				SOCK_SEQPACKET,
				(struct sockaddr *)&server_addr,
				&our_conn_params);
	TRACE_DBG(LOG_APPLICATION, "%s server: openserver returned "
			"listening_fd %d", time_prefix(), listening_fd);
	if (listening_fd < 0)
		exit(EXIT_FAILURE);

	/* we are now successfully established as a server */
	for (; ;) {

		print_ip_address((struct sockaddr *)&server_addr,
						"server listening", "latenexs");
		print_attributes(listening_fd, "server");

		/* set up the address vector so exs_accept will return the
		 * peer address of the new client
		 */
		address_vector[0].exs_addr=(struct sockaddr*)&param.client_addr;
		address_vector[0].exs_addrlen = sizeof(param.client_addr);
		address_vector[0].exs_ahandle = NULL;

		TRACE_DBG(LOG_APPLICATION,
				"%s server: exs_addr %p, exs_addrlen %d,"
				" exs_ahandle %p", time_prefix(),
				address_vector[0].exs_addr,
				address_vector[0].exs_addrlen,
				address_vector[0].exs_ahandle);

		/* accept a client connection (block until one arrives) */
		if (exs_accept(listening_fd, address_vector, ADDRESS_ITEMS, 0,
				accept_queue) < 0 ) {
			perror("server: exs_accept");
			exit(EXIT_FAILURE);
		}
		TRACE_DBG(LOG_APPLICATION, "%s server: exs_accept returned ok",
				time_prefix());

		/* in exs need to wait for accept to complete asynchronously */
		nevents = exs_qdequeue(accept_queue,accept_vector,ACCEPT_ITEMS,
					NULL);
		TRACE_DBG(LOG_APPLICATION, "%s server: exs_qdequeue nevents = %d "
				"accept_vector[0].exs_evt_type = %d",
				time_prefix(), nevents,
				accept_vector[0].exs_evt_type);
		if (nevents < 0) {
			perror("server: exs_qdequeue");
			exit(EXIT_FAILURE);
		}
		if (nevents != 1) {
			fprintf(stderr, "%s server: exs_qdequeue dequeued %d "
					"items, expected 1\n",
					time_prefix(), nevents);
			exit(EXIT_FAILURE);
		}
		if (accept_vector[0].exs_evt_type != EXS_EVT_ACCEPT) {
			fprintf(stderr, "%s server: exs_qdequeue dequeued item "
					"type %d, expected %d\n", time_prefix(),
					accept_vector[0].exs_evt_type,
					EXS_EVT_ACCEPT);
			exit(EXIT_FAILURE);
		}
		TRACE_DBG(LOG_APPLICATION,"%s server: nevents = %d, exs_evt_type ="
				" EXS_EVT_ACCEPT", time_prefix(), nevents);
		if (accept_vector[0].exs_evt_errno != 0) {
			errno = accept_vector[0].exs_evt_errno;
			perror("server: exs_qdequeue status");
			exit(EXIT_FAILURE);
		}
		TRACE_DBG(LOG_APPLICATION, "%s server: exs_evt_errno == 0",
				time_prefix());
		param.client_fd = accept_vector[0].exs_evt_union.exs_evt_accept
				.exs_evt_new_socket;
		TRACE_DBG(LOG_APPLICATION, "%s server: exs_evt_new_socket %d, "
				"new_client %p", time_prefix(),
				param.client_fd, &param.client_addr);

		/* When exs becomes asynchronous, this can be threaded.  Not now */
		//errno = pthread_create(&kid_id, NULL, server_thread, &param);
		//if (errno != 0) {
		//	perror("server: pthread_create");
		//} else {
		//	pthread_detach(kid_id);
		//	fprintf(stderr, "*** child with thread_id %lu started\n",
		//		(long)kid_id);
		//}
		/* Here is the synchronous version */
		server_thread(&param);

		/* For now, go through this loop only once.  Future versions of
		 * this program will loop infinitely. */
		break;
	}

	if (exs_qdelete(accept_queue) < 0) {
		perror("server: exs_qdelete");
	}

	/* close the "listening post" socket by which server made connections */
	TRACE_DBG(LOG_APPLICATION, "%s server: closing listening post %d",
			time_prefix(), listening_fd);
	if (exs_blocking_close(listening_fd) < 0) {
		perror("server: exs_blocking_close listening post");
	}

	exit(EXIT_SUCCESS);
}	/* main */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
