/*	evt_queue.c	*/

/*
 * Copyright 2007, 2008, 2009, 2010, 2011, 2012, 2013, 2014 by
 *	University of New Hampshire Interoperability Laboratory
 *	121 Technology Drive, Suite 2
 *	Durham, NH 03824-4716
 *	phone: 1-603-862-0090
 *	fax: 1-603-862-4181
 *	e-mail: info@iol.unh.edu
 * All rights reserved.
 *
 * This file is part of the UNH EXS implementation of the
 * Consortium Specification: Extended Sockets API (ES-API) Issue 1.0,
 * Document Number C050 published by The Open Group, January 2005,
 * and copyright 2005 by The Open Group.
 *
 * This software is available to you under a choice of one of two
 * licenses.  You may choose to be licensed under the terms of the GNU
 * General Public License (GPL) Version 2, available from the file
 * COPYING.GPL in the main directory of this source tree, or the
 * BSD license, available from the file COPYING.BSD in the main directory
 * of this source tree.
 *
 * This material is based upon work supported by the National Science
 * Foundation under Grant No. OCI-1127228.
 *
 * Any opinions, findings, and conclusions or recommendations
 * expressed in this material are those of the authors and do not
 * necessarily reflect the views of the National Science Foundation.
 */

/*
 * The UNH IOL implementation of the Extended Sockets API (ES-API) Issue 1.0
 *
 * defined in:	The Interconnect Software Consortium Issue 1.0
 *		ISBN: 1-931624-52-6
 *		Published by The open Group, January 2005.
 *
 * all page references in this source code are to pages in that document.
 */

/*
 * Most of the coding in this file done by:
 *	Bob Russell
 *	rdr@iol.unh.edu
 */

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#define _POSIX_C_SOURCE 200112L
#define _ISOC99_SOURCE
#define _XOPEN_SOURCE 600

#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#include <string.h>
#include <errno.h>
#include <semaphore.h>
#include <signal.h>
#include <time.h>
#include <pthread.h>
#include <limits.h>

#include <atomic.h>
#include <evt_queue.h>
#include <our_calloc_free.h>

/* default "depth" of event queue */
#define OUR_DEFAULT_QUEUE_DEPTH 64

/* number of slots per (re)allocation in queue_table */
#define	QUEUE_CHUNK_SIZE	64

/* table of qhandles that point to all created struct evt_queue */
static our_qhandle_t *queue_table = NULL;

/* number of slots currently allocated to queue_table */
static unsigned long queue_table_limit = 0;

/* lowest index to free slot in queue_table (0 unused) */
static unsigned long lowest_available_qnumb = 1;

/* protect all access to queue_table */
static pthread_mutex_t	qt_mutex = PTHREAD_MUTEX_INITIALIZER;


/* called locally by _evt_queue_init(), _swap_evt_to_full_evt_queue(),
 *		and _post_event_to_full_evt_queue()
 * all these are local, so qhandle is valid and qt_mutex is already locked
 */
static exs_event_t *
alloc_exs_event( our_qhandle_t qhandle )
{
	exs_event_t	*evt_ptr;

	evt_ptr = EXS_CALLOC(1, sizeof(struct exs_event));
	if (evt_ptr == NULL)
	{
		OUR_PERROR("alloc_exs_event", errno);
		return NULL;
	}
	exs_p_list_add_tail(&qhandle->empty_list, evt_ptr);
	TRACE_DBG(EXS_LOG_QUEUE,
		"qhandle %p add new empty event %p to empty_list %p",
		qhandle, evt_ptr, &qhandle->empty_list);
	return evt_ptr;
}	/* alloc_exs_event */


/* create and initialize a new event queue capable of handling "depth" events
 * Called from our_get_from_queue_cache(), exs_qcreate(), exs_blocking_close()
 *
 * Returns	pointer to the new queue on success
 *		NULL on failure, in which case global errno has error code
 */
our_qhandle_t
_evt_queue_init( unsigned int depth )
{
	int			ret;
	our_qhandle_t		qhandle, *newtable;
	unsigned long		new_qnumb, newlimit;
	struct exs_event	*evt_ptr;

	TRACE_DBG(EXS_LOG_QUEUE,
		"start, queue_table %p size %lu lowest %lu",
		queue_table, queue_table_limit,
		lowest_available_qnumb);

	/* allocate the evt_queue structure */
	qhandle = EXS_CALLOC(1, sizeof(struct evt_queue));
	if (qhandle == NULL)
		return NULL;

	TRACE_DBG(EXS_LOG_QUEUE,
		"allocated qhandle %p, %zu bytes",
		qhandle, sizeof(struct evt_queue));

	ret = pthread_cond_init(&qhandle->get_out_notify, NULL);
	if (ret != 0) {
		EXS_FREE(qhandle);
		return NULL;
	}

	/* initialize the full list variables */
	exs_p_list_init(&qhandle->full_list, OUR_LIST_BLOCKING);

	/* initialize the empty list variables */
	exs_p_list_init(&qhandle->empty_list, 0);

	if (depth == 0) {
		/* user wants use to set a default depth */
		depth = OUR_DEFAULT_QUEUE_DEPTH;
	}

	qhandle->depth = qhandle->nitems = depth;

	TRACE_DBG(EXS_LOG_QUEUE,
		"set qhandle %p depth %u",
		qhandle, depth);

	/* allocate all the "struct exs_event" objects on this queue */
	while (depth > 0) {
		if (alloc_exs_event(qhandle) == NULL)
		{
			ret = errno;
			goto err3;
		}
		depth--;
	}

	/* now add this newly-constructed qhandle to queue table */
	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	for (new_qnumb = lowest_available_qnumb; new_qnumb < queue_table_limit;
						new_qnumb++) {
		if (queue_table[new_qnumb] == NULL) {
			/* found a free slot, use it */
			TRACE_DBG(EXS_LOG_QUEUE,
				"found free slot %lu between lowest %lu and limit %lu",
				new_qnumb, lowest_available_qnumb,
				queue_table_limit);
			break;
		}
	}	/* for */
	if (new_qnumb >= queue_table_limit) {
		/* table is full, try to increase its size */
		TRACE_DBG(EXS_LOG_QUEUE,
			"queue_table full, old base %p, old limit %lu",
			queue_table, queue_table_limit);

		/* if queue_table is NULL, realloc == alloc,
		 *	new_qnumb = lowest = 1, limit = 0;
		 * else new_qnumb = limit = k*CHUNK_SIZE
		 */
		newlimit = queue_table_limit + QUEUE_CHUNK_SIZE;
		newtable = realloc(queue_table,newlimit*sizeof(our_qhandle_t));
		if (newtable == NULL) {
			/* can't increase size, don't create new queue */
			ret = ENOMEM;
		} else {
			/* NULL out new space just appended to end of table */
			memset(&newtable[queue_table_limit], 0,
				QUEUE_CHUNK_SIZE * sizeof(our_qhandle_t));

			/* update static global table parameters */
			queue_table = newtable;
			queue_table_limit = newlimit;
			TRACE_DBG(EXS_LOG_QUEUE,
				"(re)alloc ok, new queue_table %p new limit %lu",
				queue_table, queue_table_limit);
		}
	}
	if (ret == 0) {
		/* new_qnumb is lowest index in queue_table_limit of free slot*/
		qhandle->qnumb = new_qnumb;
		queue_table[new_qnumb] = qhandle;
		TRACE_DBG(EXS_LOG_QUEUE,
			"queue_table[%lu] %p",
			new_qnumb, qhandle);
		/* update lowest_available_qnumb
		 * (which may force it off current end of table)
		 */
		lowest_available_qnumb = new_qnumb + 1;
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	if (ret == 0) {
		/* successful, return pointer to the new "struct evt_queue"*/
		TRACE_DBG(EXS_LOG_QUEUE,
			"leave, queue_table %p limit %lu lowest %lu qnumb %lu ret %p",
			queue_table,
			queue_table_limit, lowest_available_qnumb,
			new_qnumb, qhandle);
		return qhandle;
	}

	/* else fall through into error recovery */
err3:
	while ((evt_ptr = exs_p_list_del_head(&qhandle->empty_list))) {
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p deallocating empty event %p",
			qhandle, evt_ptr);
		EXS_FREE(evt_ptr);
	}

	pthread_cond_destroy(&qhandle->get_out_notify);
	EXS_FREE(qhandle);
	errno = ret;
	return NULL;
}	/* _evt_queue_init */


/* called from our_destroy_queue_cache(), exs_qdelete(), exs_blocking_close()
 * returns 0 if all ok, -1 on error with global errno set
 */
int
_evt_queue_deinit( our_qhandle_t qhandle )
{
	int			ret;
	unsigned long		qnumb;
	struct exs_event	*evt_ptr;


	TRACE_DBG(EXS_LOG_QUEUE, "qhandle %p", qhandle);

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else if (qhandle->npending > 0) {
		ret = -1;
		errno = EBUSY;
	} else {

		/* we are the first to shut this down, mark it */
		qhandle->get_out_now = 1;
		while(qhandle->waiting_for_notify) {
			TRACE_DBG(EXS_LOG_SYNC,
				"WAIT get_out_notify %p, qt_mutex %p",
				&qhandle->get_out_notify, &qt_mutex);

			pthread_cond_wait(&qhandle->get_out_notify, &qt_mutex);

			TRACE_DBG(EXS_LOG_SYNC,
				"AWAKE get_out_notify %p, qt_mutex %p",
				&qhandle->get_out_notify, &qt_mutex);
		}	/* while */

		/* ok, NULL out the table and update lowest available */
		queue_table[qnumb] = NULL;
		if (lowest_available_qnumb > qnumb)
			lowest_available_qnumb = qnumb;
		ret = 0;
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	if (ret) {
		/* qhandle is not valid, don't try to use it */
		goto out;
	}

	/* qhandle is valid and has been removed from queue table
	 * free up its parts and then free it
	 */
	while ((evt_ptr = exs_p_list_del_head(&qhandle->full_list))) {
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p deallocating %zu-byte event %p type %d",
			qhandle, sizeof(exs_event_t), evt_ptr,
			evt_ptr->exs_evt_type);
		EXS_FREE(evt_ptr);
	}
	exs_p_list_destroy(&qhandle->full_list);

	while ((evt_ptr = exs_p_list_del_head(&qhandle->empty_list))) {
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p deallocating empty event %p",
			qhandle, evt_ptr);
		EXS_FREE(evt_ptr);
	}
	exs_p_list_destroy(&qhandle->empty_list);

	pthread_cond_destroy(&qhandle->get_out_notify);
	EXS_FREE(qhandle);

out:
	TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p returning %d",
			qhandle, ret);

	return ret;
}	/* _evt_queue_deinit */


/* disables signal generation for this queue
 * returns 0 if all ok, -1 with global errno set on any error
 */
int
_evt_queue_disable_signal( our_qhandle_t qhandle )
{
	int			ret;
	unsigned long		qnumb;

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		qhandle->signal_number = 0;
		ret = 0;
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	return ret;
}	/* _evt_queue_disable_signal */


/* enables signal generation for this queue whenever a new event is
 * enqueued on a previously empty queue
 * returns 0 if all ok, -1 and errno set on any error
 */
int
_evt_queue_enable_signal( our_qhandle_t qhandle, int signal_number )
{
	int			ret, sig;
	unsigned long		qnumb;

	if (signal_number <= 0) {
		errno = EINVAL;
		return -1;
	}

	sig = 0;	/* assume we do not send any signal */

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		qhandle->signal_number = signal_number;
		/* also raise the signal if there are currently
		 * events on this queue
		 */
		if(!exs_p_list_empty(&qhandle->full_list)) {
			sig = signal_number;
		}
		ret = 0;
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	if (ret == 0 && sig != 0) {
		ret = raise(sig);
	}
	return ret;
}	/* _evt_queue_enable_signal */


/* Called from exs_qmodify()
 *
 * Raises any non-zero signo
 * returns 0 if all ok, else -1 with errno set
 */
int
_evt_queue_do_notify( our_qhandle_t qhandle, int signo )
{
	int			ret;
	unsigned long		qnumb;

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		ret = 0;
	}

	if (ret == 0 && signo > 0) {
		/* qhandle is valid and signo is positive, raise it */
		ret = raise(signo);
	}
	return ret;
}	/* _evt_queue_do_notify */


/* returns in vector up to count events posted to qhandle
 * returns the number actually copied into vector if all ok
 * returns -1 with errno set if any error
 *
 * Called from exs_qdequeue(), exs_accept(), exs_connect(), exs_send(),
 *		exs_recv(), exs_blocking_close()
 * caller gives us timeout in struct timeval (seconds and microseconds);
 * our clock function and timedwait use struct timespec (seconds and nanosecs);
 */
int
_evt_queue_dequeue_vector( our_qhandle_t qhandle, exs_event_t *vector,
			int count, const struct timeval *timeout )
{
	int			n, kount;
	int			err = 0;
	unsigned int		wait_count;
	unsigned long		qnumb;
	struct exs_event	*evt_ptr;
	struct timespec		abs_delay;

	if (count <= 0) {
		errno = EINVAL;
		return -1;
	}

	/* do this as soon as arrive, before locking, to get good abs time */
	if (timeout != NULL) {
		clock_gettime(CLOCK_REALTIME, &abs_delay);
		abs_delay.tv_sec += timeout->tv_sec;
		abs_delay.tv_nsec += 1000 * timeout->tv_usec;
		if (abs_delay.tv_nsec >= 1000000000) {
			abs_delay.tv_nsec -= 1000000000;
			abs_delay.tv_sec += 1;
		}
	}

	TRACE_DBG(EXS_LOG_QUEUE,
		"qhandle %p, vector %p, count %d",
		qhandle, vector, count);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		n = -1;		/* qhandle is not valid */
		errno = EINVAL;
		goto errout;
	}

	/* qhandle is valid, see if queue has any events on it */
	n = 0;
	atomic_add_reset(1, UINT_MAX, &qhandle->waiting_for_notify);
	if (timeout) {
		if (timeout->tv_sec == 0 && timeout->tv_usec == 0) {
			/* Do a strict poll; do not go through any conditional
			 * waits, even with a delay of 0. */
			errno = 0;
			evt_ptr = exs_p_list_del_head(&qhandle->full_list);
		} else {
			evt_ptr = exs_p_list_del_head_or_wait(&qhandle->full_list,
								&abs_delay);
		}
	} else {
		evt_ptr = exs_p_list_del_head_or_wait(&qhandle->full_list,NULL);
	}
	if (!evt_ptr) {
		err = errno;
	}
	wait_count=atomic_add_reset(-1, UINT_MAX, &qhandle->waiting_for_notify);

	/* qhandle may have changed while we were waiting on an empty
	 * queue, recheck that qhandle is still valid */
	if (queue_table[qnumb] != qhandle) {
		n = -1;		/* qhandle is not valid */
		errno = EINVAL;
		goto errout;
	} else if (qhandle->get_out_now) {
		/* someone else is shutting this queue
		 * down, leave at once
		 */
		if (wait_count == 0) {
			/* we are the last ones out */
			TRACE_DBG(EXS_LOG_SYNC,
					"BROADCAST get_out_notify %p",
					&qhandle->get_out_notify);
			pthread_cond_broadcast(
				     &qhandle->get_out_notify);
			}
		errno = EINVAL;
		n = -1;
		goto errout;
	}

	if (!evt_ptr) {
		if (err == 0 || err == ETIMEDOUT) {
			errno = 0;
			n = 0;
		} else {
			errno = err;
			n = -1;
		}
		goto errout;
	} else {
		memcpy(&vector[0], evt_ptr, sizeof(exs_event_t));
		n++;
		exs_p_list_add_tail(&qhandle->empty_list, evt_ptr);
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p move %zu-byte event %p type %d from full to empty list",
			qhandle, sizeof(exs_event_t),
			evt_ptr, evt_ptr->exs_evt_type);
	}

	/* ret == 0, so should be an event on the queue now,
	 * return up to "count" of them */
	kount = count;
	while (--kount > 0 && (evt_ptr = exs_p_list_del_head(
						&qhandle->full_list))) {
		memcpy(&vector[n], evt_ptr, sizeof(exs_event_t));
		n++;
		exs_p_list_add_tail(&qhandle->empty_list, evt_ptr);
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p move %zu-byte event %p type %d from full to empty list",
			qhandle, sizeof(exs_event_t),
			evt_ptr, evt_ptr->exs_evt_type);
		if (--kount <= 0)
			break;
	}	/* list_for_each_safe */

errout:
	TRACE_DBG(EXS_LOG_QUEUE,
		"qhandle %p dequeue to vector %p returning %d",
		qhandle, vector, n);

	return n;
}	/* _evt_queue_dequeue_vector */


/* if qhandle's empty list is not itself empty,
 * then removes and returns first event struct on that list
 * else returns NULL
 *
 * Called from our_trigger_poll_notify() and exs_accept()
 */
exs_event_t *
_get_empty_evt_queue_event( our_qhandle_t qhandle, exs_qhandle_t qn )
{
	exs_event_t		*evt_ptr;


	TRACE_DBG(EXS_LOG_QUEUE,
		"qhandle %p qn %p queue_table[qn] %p",
		qhandle, qn, queue_table[(unsigned long)qn]);

	/* check that the (qhandle, qn) pair is still valid */
	if (queue_table[(unsigned long)qn] != qhandle
				|| qhandle->qnumb != (unsigned long)qn
				|| qhandle->get_out_now != 0) {
		/* qhandle is not valid */
		errno = EINVAL;
		evt_ptr = NULL;
	} else {
		/* qhandle is valid, use it */
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p qn %p empty_list %p empty %d",
			qhandle, qn, &qhandle->empty_list,
			exs_p_list_empty(&qhandle->empty_list));
		if (!(evt_ptr = exs_p_list_del_head(&qhandle->empty_list))) {
			/* no empty event structs, return error immediately */
			/* FIXME: try to make a new one first */
			errno = ENOMEM;		/* find a better errno? */
		}
	}

	TRACE_DBG(EXS_LOG_QUEUE,
		"qhandle %p remove empty event %p from empty_list",
		qhandle, evt_ptr);
	return evt_ptr;
}	/* _get_empty_evt_queue_event */


/* adds evt_ptr to end of qhandle
 * returns 0 if all ok, else -1 with errno set
 *
 * Called from our_trigger_poll_notify() and our_fill_in_accept_event()
 */
int
_put_full_evt_queue_event( our_qhandle_t qhandle, exs_event_t *evt_ptr )
{
	int			signo = 0;
	int			ret;
	unsigned long		qnumb;


	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		ret = 0;
		if (qhandle->signal_number > 0
			&& exs_p_list_empty(&qhandle->full_list)){
			/* adding first event onto empty queue
			* need to send signal to this process
			*/
			signo = qhandle->signal_number;
		}
		TRACE_DBG(EXS_LOG_QUEUE,
			"qhandle %p add full event %p to full_list, signo %d",
			qhandle, evt_ptr, signo);

		exs_p_list_add_tail(&qhandle->full_list, evt_ptr);
		atomic_add_reset(-1, UINT_MAX, &qhandle->npending);
	}

	if (signo > 0) {
		/* qhandle was valid and signo is positive, raise it */
		ret = raise(signo);
	}
	return ret;
}	/* _put_full_evt_queue_event */


/* Called by our_post_user_event() and our_find_send_to_match().
 *
 * 1) add the exs_event structure pointed to by *evt_ptr
 *    to the end of the qhandle's full_list
 * 2) remove the exs_event structure from the front of the qhandle's empty_list
 *    (if empty_list is itself empty, allocate a new exs_event structure)
 *    and set *evt_ptr to point to it
 *
 * Caller must check if *evt_ptr is NULL, and if it is, then a new event
 * structure could not be allocated!
 *
 * Returns signo >= 0 to force
 * our_post_user_event to broadcast the notify and send the signal.
 */
int
_swap_evt_to_full_evt_queue( our_qhandle_t qhandle, exs_event_t **evt_ptr )
{
	exs_event_t		*tmp_evt_ptr;
	int			signo = 0;
	unsigned long		qnumb;


	if (qhandle == NULL) {
		/* user does NOT want to get a notify queued up */
		return 0;
	}

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		/* qhandle is not valid */
		errno = EINVAL;
		signo = -1;
	} else {
		/* qhandle is valid, use it */

		if (qhandle->signal_number > 0
			&& exs_p_list_empty(&qhandle->full_list)){
			/* adding first event onto full_list,
			* so need to send signal to this process
			*/
			signo = qhandle->signal_number;
		}

		/* Grab a copy of the caller's current event pointer */
		tmp_evt_ptr = *evt_ptr;

		/* Replace the caller's event pointer with the new one */
		if (!(*evt_ptr = exs_p_list_del_head(&qhandle->empty_list))) {
			/* Note: This could return NULL and caller must deal
			 * with the result */
			*evt_ptr = alloc_exs_event(qhandle);
		}
		TRACE_DBG(EXS_LOG_QUEUE,
			"remove exs_event %p from empty list %p",
			*evt_ptr, &qhandle->empty_list);

		/* Now post the original caller's event */
		TRACE_DBG(EXS_LOG_QUEUE,
			"add event %p to full list %p, signo %d",
			*evt_ptr, &qhandle->full_list, signo);
		exs_p_list_add_tail(&qhandle->full_list, tmp_evt_ptr);

		atomic_add_reset(-1, UINT_MAX, &qhandle->npending);
	}

	return signo;
}	/* _swap_evt_to_full_evt_queue */


/*
 * on entry, qhandle is known to be not NULL.
 * on entry, conn_pending->qhandle is known to be not NULL.
 * on exit, conn_pending->qhandle is always set to be NULL.
 *
 * Called from our_connect_returning(), our_undo_exs_socket(),
 *		and our_close_accepting()
 *
 * if all ok, gets a new exs_evt_t item from qhandle's empty list,
 * fills it in with event info, and returns in qhandle and signo
 * the info necessary for caller to post this event
 *
 * returns 0 if all ok, else -1 with errno set (although result is never used!)
 */
int
_post_event_to_full_evt_queue( struct pending_connect *conn_pending,
				our_qhandle_t *qhandle, int *signo )
{
	int			ret;
	unsigned long		qnumb;
	exs_event_t		*evt_ptr;

	*signo = 0;

	/* check that qhandle is still valid */
	if ((qnumb = conn_pending->qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| conn_pending->qhandle->get_out_now != 0
				|| queue_table[qnumb]!=conn_pending->qhandle) {
		ret = -1;	/* qhandle is not valid */
		*qhandle = NULL;
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		ret = 0;

		*qhandle = conn_pending->qhandle;

		/* once event is posted, the qhandle is used up */
		conn_pending->qhandle = NULL;

		if ((*qhandle)->signal_number > 0
		&& exs_p_list_empty(&(*qhandle)->full_list)) {
			/* adding first event onto empty queue
			* need to send signal to this process
			*/
			*signo = (*qhandle)->signal_number;
		}

		if (!(evt_ptr = exs_p_list_del_head(&(*qhandle)->empty_list))) {
			if((evt_ptr = alloc_exs_event(*qhandle)) == NULL) {
				/* errno already set to ENOMEM
				* event is never sent -- lost!
				*/
				*signo = -1;	/* don't signal any event */
				ret = -1;
				goto err1;
			}
		}

		/* post new connect event in first item on empty list */
		/* see page 18 for these settings */

		evt_ptr->exs_evt_type = conn_pending->evt_type;
		evt_ptr->exs_evt_errno = conn_pending->err_numb;
		evt_ptr->exs_evt_ahandle = conn_pending->ahandle;
		evt_ptr->exs_evt_socket = conn_pending->sockno;

		/* move the (now full) event from empty to full list */
		TRACE_DBG(EXS_LOG_QUEUE,
			"move exs_event %p type %d from empty list %p to full list %p",
			evt_ptr, evt_ptr->exs_evt_type,
			&(*qhandle)->empty_list, &(*qhandle)->full_list);

		exs_p_list_add_tail(&(*qhandle)->full_list, evt_ptr);
		atomic_add_reset(-1, UINT_MAX, &(*qhandle)->npending);
		ret = 0;
	}
err1:

	/* leave without actually broadcasting the post or raising the signal */
	return ret;
}	/* _post_event_to_full_evt_queue */


/* Called only from exs_qstatus()
 * Returns 0 if all ok and sets result, else -1 with errno set
 */
int
_get_evt_queue_depth( our_qhandle_t qhandle, int *result )
{
	int			ret;
	unsigned long		qnumb;

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		*result = qhandle->depth;
		ret = 0;
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	return ret;
}	/* _get_evt_queue_depth */


/* Called only from exs_qstatus()
 * Returns 0 if all ok and sets sig, else -1 with errno set
 */
int
_get_evt_queue_signal( our_qhandle_t qhandle, struct exs_signal *sig)
{
	int			ret;
	unsigned long		qnumb;

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		ret = -1;	/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		ret = 0;
		if (qhandle->signal_number > 0) {
			sig->exs_signo=qhandle->signal_number;
			sig->exs_sigstate = EXS_SIG_ENABLE;
		} else {
			sig->exs_signo = 0;
			sig->exs_sigstate = EXS_SIG_DISABLE;
		}
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	return ret;
}	/* _get_evt_queue_signal */


/* if all ok, returns 0
 *	and sets in result number of events currently on qhandle's full_list
 * else returns -1 with errno set
 *
 * Called only from exs_qstatus()
 */
int
_get_evt_queue_n_events( our_qhandle_t qhandle, int *result )
{
	int			n;
	unsigned long		qnumb;

	TRACE_DBG(EXS_LOG_SYNC, "LOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_lock(&qt_mutex);
	TRACE_DBG(EXS_LOG_SYNC, "PAST qt_mutex %p", &qt_mutex);

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		n = -1;		/* qhandle is not valid */
		errno = EINVAL;
	} else {
		/* qhandle is valid, use it */
		*result = exs_p_list_size(&qhandle->full_list);
	}

	TRACE_DBG(EXS_LOG_SYNC, "UNLOCK qt_mutex %p", &qt_mutex);
	pthread_mutex_unlock(&qt_mutex);

	return n;
}	/* _get_evt_queue_n_events */


/* adds n to the pending event counter of this queue
 * returns 0 if all ok, -1 with global errno set on any error
 */
int
_evt_queue_add_pending(our_qhandle_t qhandle, unsigned int n)
{
	unsigned long		qnumb;
	unsigned int		new_npending;

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		errno = EINVAL;
		return -1;
	}

	/* qhandle is valid, use it */
	new_npending = atomic_add_reset(n, UINT_MAX, &qhandle->npending);
	return 0;
}	/* _evt_queue_add_pending */


/* removes n from the pending event counter of this queue
 * returns 0 if all ok, -1 with global errno set on any error
 */
int
_evt_queue_rm_pending(our_qhandle_t qhandle, int n)
{
	unsigned long		qnumb;
	unsigned int		new_npending;

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		errno = EINVAL;
		return -1;
	}

	/* qhandle is valid, use it */
	new_npending = atomic_add(-(int)n, &qhandle->npending);
	return 0;
}	/* _evt_queue_add_pending */


/* basic conversion from user's qnumb to our qhandle format
 * Called near the beginning of most exs_xxx() functions
 * to verify that the user's qhandle parameter is valid.
 *
 * Returns true pointer to struct evt_queue if ok, else NULL and errno is set
 *
 * Designed to return NULL (error) in all 3 special cases:
 *	qh == NULL
 *	qh == EXS_QHANDLE_INVALID
 *	qh == EXS_QHANDLE_BLOCKING
 */
our_qhandle_t
_cnv_qnumb_to_ptr( exs_qhandle_t qn )
{
	our_qhandle_t	qhandle;
	unsigned long	qnumb;

	/* check that user's qn is valid */
	if ((qnumb = (unsigned long)qn) >= queue_table_limit
				|| (qhandle = queue_table[qnumb]) == NULL
				|| qhandle->get_out_now != 0
				|| qhandle->qnumb != qnumb) {
		qhandle = NULL;	/* qnumb is not valid */
		errno = EINVAL;	/* Invalid argument */
	}

	TRACE_DBG(EXS_LOG_QUEUE,
		"exs_qhandle %p return our_qhandle %p",
		qn, qhandle);
	return qhandle;
}	/* _cnv_qnumb_to_ptr */

/* basic conversion from our qhandle format to user's exs_qhandle_t
 * Called only from exs_qcreate() and exs_blocking_close()
 *
 * Returns user's pseudo-pointer if ok, else NULL (and errno is NOT set)
 */
exs_qhandle_t
_cnv_ptr_to_qnumb( our_qhandle_t qhandle )
{
	unsigned long	qnumb;

	/* check that qhandle is still valid */
	if ((qnumb = qhandle->qnumb) <= 0
				|| qnumb >= queue_table_limit
				|| qhandle->get_out_now != 0
				|| queue_table[qnumb] != qhandle) {
		qnumb = 0;	/* qhandle is not valid */
	}

	TRACE_DBG(EXS_LOG_QUEUE,
		"our_qhandle %p return exs_qhandle %p",
		qhandle, (exs_qhandle_t)qnumb);
	return (exs_qhandle_t)qnumb;
}	/* _cnv_ptr_to_qnumb */

/* vi: set autoindent tabstop=8 shiftwidth=8 : */
